package avlindex

import (
	"bytes"
	"encoding/gob"
	"io/ioutil"
	"log"
	"os"
	"testing"

	"github.com/huandu/skiplist"

	"github.com/Pallinder/go-randomdata"
	"github.com/davecgh/go-spew/spew"
	"github.com/emirpasic/gods/trees/avltree"
	"github.com/emirpasic/gods/trees/redblacktree"
	"github.com/emirpasic/gods/utils"
)

const CompartorSize = 1000000
const NumberMax = 50000000

func TestSave(t *testing.T) {

	f, err := os.OpenFile("../l.log", os.O_CREATE|os.O_TRUNC|os.O_WRONLY, 0666)
	if err != nil {
		log.Println(err)
	}

	//fmt.Println(userBytes)

	var l []int

	// for i := 0; len(l) < 1000; i++ {
	// 	v := randomdata.Number(0, 65535)
	// 	l = append(l, v)
	// }

	//m := make(map[int]int)
	for i := 0; len(l) < CompartorSize; i++ {
		v := randomdata.Number(0, NumberMax)
		// if _, ok := m[v]; !ok {
		// 	m[v] = v
		l = append(l, v)
		// }
	}

	var result bytes.Buffer
	encoder := gob.NewEncoder(&result)
	encoder.Encode(l)
	lbytes := result.Bytes()
	f.Write(lbytes)

}

func loadTestData() []int {
	data, err := ioutil.ReadFile("../l.log")
	if err != nil {
		log.Println(err)
	}
	var l []int
	decoder := gob.NewDecoder(bytes.NewReader(data))
	decoder.Decode(&l)
	return l
}

func TestIterator(t *testing.T) {
	// avl := New(utils.IntComparator)
	// for _, v := range []int{7, 14, 14, 14, 16, 17, 20, 30, 21, 40, 40, 50, 3, 40, 40, 40, 15} {
	// 	avl.Put(v)
	// }
	// t.Error(avl.Values())
	// t.Error(avl.debugString())

}

func TestGetRange(t *testing.T) {
	tree := New(utils.IntComparator)
	l := []int{7, 14, 14, 14, 16, 17, 20, 30, 21, 40, 50, 3, 40, 40, 40, 15}
	for _, v := range l {
		tree.Put(v)
	}
	// [3 7 14 14 14 15 16 17 20 21 30 40 40 40 40 50]
	// t.Error(tree.Values(), tree.Size())

	var result string
	result = spew.Sprint(tree.GetRange(0, 5))
	if result != "[3 7 14 14 14 15] true" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(2, 5))
	if result != "[14 14 14 15] true" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(10, 100))
	if result != "[30 40 40 40 40 50] false" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(15, 0)) // size = 16, index max = 15
	if result != "[50 40 40 40 40 30 21 20 17 16 15 14 14 14 7 3] true" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(16, 0)) // size = 16, index max = 15
	if result != "[50 40 40 40 40 30 21 20 17 16 15 14 14 14 7 3] false" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(5, 1)) // size = 16, index max = 15
	if result != "[15 14 14 14 7] true" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(-1, -5)) // size = 16, index max = 15
	if result != "[50 40 40 40 40] true" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(-1, -16)) // size = 16, index max = 0 - 15 (-1,-16)
	if result != "[50 40 40 40 40 30 21 20 17 16 15 14 14 14 7 3] true" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(-1, -17)) // size = 16, index max = 0 - 15 (-1,-16)
	if result != "[50 40 40 40 40 30 21 20 17 16 15 14 14 14 7 3] false" {
		t.Error(result)
	}

	result = spew.Sprint(tree.GetRange(-5, -1)) // size = 16, index max = 0 - 15 (-1,-16)
	if result != "[40 40 40 40 50] true" {
		t.Error(result)
	}
}

func TestGetAround(t *testing.T) {
	avl := New(utils.IntComparator)
	for _, v := range []int{7, 14, 14, 14, 16, 17, 20, 30, 21, 40, 50, 3, 40, 40, 40, 15} {
		avl.Put(v)
	}

	var Result string

	Result = spew.Sprint(avl.GetAround(3))
	if Result != "[7 3 <nil>]" {
		t.Error("avl.GetAround(3)) is error", Result)
	}

	Result = spew.Sprint(avl.GetAround(40))
	if Result != "[40 40 30]" {
		t.Error("avl.GetAround(40)) is error", Result)
	}

	Result = spew.Sprint(avl.GetAround(50))
	if Result != "[<nil> 50 40]" {
		t.Error("avl.GetAround(50)) is error", Result)
	}
}

// // for test error case

func TestPutComparatorRandom(t *testing.T) {

	for n := 0; n < 300000; n++ {
		avl := New(utils.IntComparator)
		godsavl := avltree.NewWithIntComparator()

		content := ""
		m := make(map[int]int)
		for i := 0; len(m) < 10; i++ {
			v := randomdata.Number(0, 65535)
			if _, ok := m[v]; !ok {
				m[v] = v
				content += spew.Sprint(v) + ","
				avl.Put(v)
				godsavl.Put(v, v)
			}
		}

		s1 := spew.Sprint(avl.Values())
		s2 := spew.Sprint(godsavl.Values())

		if s1 != s2 {
			t.Error(godsavl.String())
			t.Error(avl.debugString())
			t.Error(content, n)
			break
		}
	}
}

func TestGet(t *testing.T) {
	avl := New(utils.IntComparator)
	for _, v := range []int{2383, 7666, 3055, 39016, 57092, 27897, 36513, 1562, 22574, 23202} {
		avl.Put(v)
	}

	for _, v := range []int{2383, 7666, 3055, 39016, 57092, 27897, 36513, 1562, 22574, 23202} {
		v, ok := avl.Get(v)
		if !ok {
			t.Error("the val not found ", v)
		}
	}

	if v, ok := avl.Get(10000); ok {
		t.Error("the val(1000) is not in tree, but is found", v)
	}
}

func TestTravalsal(t *testing.T) {
	tree := New(utils.IntComparator)

	l := loadTestData()
	N := len(l)
	for i := 0; i < N; i++ {
		tree.Put(l[i])
	}

	i := 0
	var result []interface{}
	tree.Traversal(func(v interface{}) bool {
		result = append(result, v)
		i++
		if i >= 10 {
			return false
		}
		return true
	})
	t.Error(result)
}

func TestRemoveAll(t *testing.T) {
ALL:
	for c := 0; c < 5000; c++ {
		avl := New(utils.IntComparator)
		gods := avltree.NewWithIntComparator()
		var l []int
		m := make(map[int]int)

		for i := 0; len(l) < 10; i++ {
			v := randomdata.Number(0, 100000)
			if _, ok := m[v]; !ok {
				m[v] = v
				l = append(l, v)
				avl.Put(v)
				gods.Put(v, v)
			}
		}

		for i := 0; i < 10; i++ {

			avl.Remove(l[i])
			gods.Remove(l[i])

			s1 := spew.Sprint(avl.Values())
			s2 := spew.Sprint(gods.Values())
			if s1 != s2 {
				t.Error("avl remove error", "avlsize = ", avl.Size())
				t.Error(avl.root, i, l[i])
				t.Error(s1)
				t.Error(s2)
				break ALL
			}
		}
	}
}

func TestRemove(t *testing.T) {

ALL:
	for N := 0; N < 50000; N++ {
		avl := New(utils.IntComparator)
		gods := avltree.NewWithIntComparator()

		var l []int
		m := make(map[int]int)

		for i := 0; len(l) < 10; i++ {
			v := randomdata.Number(0, 100)
			if _, ok := m[v]; !ok {
				l = append(l, v)
				m[v] = v
				avl.Put(v)
				gods.Put(v, v)
			}
		}

		src1 := avl.String()
		src2 := gods.String()

		for i := 0; i < 10; i++ {
			avl.Remove(l[i])
			gods.Remove(l[i])
			if avl.root != nil && spew.Sprint(gods.Values()) != spew.Sprint(avl.Values()) {
				// if gods.String() != avl.String() && gods.Size() != 0 && avl.size != 0 {
				t.Error(src1)
				t.Error(src2)
				t.Error(avl.debugString())
				t.Error(gods.String())
				t.Error(l[i])
				// t.Error(avl.TraversalDepth(-1))
				// t.Error(gods.Values())
				break ALL
			}
		}
	}
}

func BenchmarkSkipListGet(b *testing.B) {
	sl := skiplist.New(skiplist.Int)
	l := loadTestData()
	b.N = len(l)

	for _, v := range l {
		sl.Set(v, v)
	}

	b.ResetTimer()
	b.StartTimer()

	execCount := 5
	b.N = len(l) * execCount

	for i := 0; i < execCount; i++ {
		for _, v := range l {
			e := sl.Get(v)
			var result [50]interface{}
			for i := 0; i < 50 && e != nil; i++ {
				result[i] = e.Value
				e = e.Next()
			}
		}
	}
}

func BenchmarkGetRange(b *testing.B) {
	tree := New(utils.IntComparator)
	l := loadTestData()
	b.N = len(l)

	for _, v := range l {
		tree.Put(v)
	}

	b.ResetTimer()
	b.StartTimer()

	execCount := 5
	b.N = len(l) * execCount

	for i := 0; i < execCount; i++ {
		for range l {
			tree.GetRange(i, i+49)
		}
	}
}

func BenchmarkSkipListSet(b *testing.B) {
	sl := skiplist.New(skiplist.Int)
	l := loadTestData()

	execCount := 50
	b.N = len(l) * execCount

	for i := 0; i < execCount; i++ {
		for _, v := range l {
			sl.Set(v, v)
		}
	}

}

func BenchmarkIterator(b *testing.B) {
	tree := New(utils.IntComparator)

	l := loadTestData()
	b.N = len(l)

	for _, v := range l {
		tree.Put(v)
	}

	b.ResetTimer()
	b.StartTimer()
	iter := tree.Iterator()
	b.N = 0
	for iter.Next() {
		b.N++
	}
	for iter.Prev() {
		b.N++
	}
	for iter.Next() {
		b.N++
	}
	b.Log(b.N, len(l))
}

func BenchmarkRemove(b *testing.B) {
	tree := New(utils.IntComparator)

	l := loadTestData()

	b.N = len(l)
	for _, v := range l {
		tree.Put(v)
	}

	b.ResetTimer()
	b.StartTimer()

	for i := 0; i < len(l); i++ {
		tree.Remove(l[i])
	}
}

func BenchmarkGodsRemove(b *testing.B) {
	tree := avltree.NewWithIntComparator()

	l := loadTestData()

	b.N = len(l)
	for _, v := range l {
		tree.Put(v, v)
	}

	b.ResetTimer()
	b.StartTimer()

	for i := 0; i < len(l); i++ {
		tree.Remove(l[i])
	}
}

func BenchmarkGodsRBRemove(b *testing.B) {
	tree := redblacktree.NewWithIntComparator()

	l := loadTestData()

	b.N = len(l)
	for _, v := range l {
		tree.Put(v, v)
	}

	b.ResetTimer()
	b.StartTimer()

	for i := 0; i < len(l); i++ {
		tree.Remove(l[i])
	}
}

func BenchmarkGet(b *testing.B) {

	tree := New(utils.IntComparator)

	l := loadTestData()
	b.N = len(l)
	for i := 0; i < b.N; i++ {
		tree.Put(l[i])
	}

	b.ResetTimer()
	b.StartTimer()

	execCount := 50
	b.N = len(l) * execCount

	for i := 0; i < execCount; i++ {
		for _, v := range l {
			tree.Get(v)
		}
	}
}

func BenchmarkGodsRBGet(b *testing.B) {
	tree := redblacktree.NewWithIntComparator()

	l := loadTestData()
	b.N = len(l)
	for i := 0; i < b.N; i++ {
		tree.Put(l[i], i)
	}

	b.ResetTimer()
	b.StartTimer()

	execCount := 50
	b.N = len(l) * execCount

	for i := 0; i < execCount; i++ {
		for _, v := range l {
			tree.Get(v)
		}
	}
}

func BenchmarkGodsAvlGet(b *testing.B) {
	tree := avltree.NewWithIntComparator()

	l := loadTestData()
	b.N = len(l)
	for i := 0; i < b.N; i++ {
		tree.Put(l[i], i)
	}

	b.ResetTimer()
	b.StartTimer()

	execCount := 50
	b.N = len(l) * execCount

	for i := 0; i < execCount; i++ {
		for _, v := range l {
			tree.Get(v)
		}
	}
}

func BenchmarkPut(b *testing.B) {
	l := loadTestData()

	b.ResetTimer()
	b.StartTimer()

	execCount := 50
	b.N = len(l) * execCount
	for i := 0; i < execCount; i++ {
		avl := New(utils.IntComparator)
		for _, v := range l {
			avl.Put(v)
		}
	}
}

func TestPutStable(t *testing.T) {

	// l := []int{14, 18, 20, 21, 22, 23, 19}
	var l []int
	for i := 0; len(l) < 10; i++ {
		l = append(l, randomdata.Number(0, 65))
	}

	avl := New(utils.IntComparator)
	for _, v := range l {
		avl.Put(v)
		t.Error(avl.debugString(), v)
	}
	t.Error(avl.Values())
	for _, v := range []int{10, 0, 9, 5, -11, -10, -1, -5} {
		t.Error(avl.Index(v))
	}

	avl.RemoveIndex(4)
	t.Error(avl.Index(4))
	t.Error(avl.Values())
	t.Error(avl.debugString())
	// t.Error(len(l), avl.debugString(), "\n", "-----------") // 3 6(4)

}

func BenchmarkIndex(b *testing.B) {
	tree := New(utils.IntComparator)

	l := loadTestData()
	b.N = len(l)
	for i := 0; i < b.N; i++ {
		tree.Put(l[i])
	}

	b.ResetTimer()
	b.StartTimer()

	b.N = 1000000

	var result [50]interface{}
	for n := 0; n < b.N; n++ {
		i := 0
		tree.Traversal(func(v interface{}) bool {
			result[i] = v
			i++
			if i < 50 {
				return true
			}
			log.Print(i)
			return false
		})
	}
}

func BenchmarkTraversal(b *testing.B) {
	tree := New(utils.IntComparator)

	l := loadTestData()
	b.N = len(l)
	for i := 0; i < b.N; i++ {
		tree.Put(l[i])
	}

	b.ResetTimer()
	b.StartTimer()

	execCount := 50
	b.N = len(l) * execCount

	for n := 0; n < execCount; n++ {
		i := 0
		var result []interface{}
		tree.Traversal(func(v interface{}) bool {
			result = append(result, v)
			i++
			if i >= 50 {
				return false
			}
			return true
		})

	}
}

func BenchmarkGodsRBPut(b *testing.B) {
	tree := redblacktree.NewWithIntComparator()

	l := loadTestData()

	b.ResetTimer()
	b.StartTimer()

	b.N = len(l)
	for _, v := range l {
		tree.Put(v, v)
	}
}

func BenchmarkGodsPut(b *testing.B) {
	tree := avltree.NewWithIntComparator()

	l := loadTestData()

	b.ResetTimer()
	b.StartTimer()

	b.N = len(l)
	for _, v := range l {
		tree.Put(v, v)
	}
}
